/*
 * ptsession.c
 *
 * Allocate/Release a pty master/slave pair, and handle utmp entries.
 *
 * Portions of code ripped from KDE's kvt:
 * Copyright (C) 1992, John Bovey, University of Kent at Canterbury,
 * with modifications by
 *	Robert Nation <nation@rocket.sanders.lockheed.com>,
 *	Garret D'Amore <garrett@netcom.com>
 *	Olaf Kirch <okir@caldera.de> (utmpenvoy handling)
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include "ptsession.h"

#ifndef UTMP_ENVOY
# define UTMP_ENVOY	"/usr/sbin/utmpenvoy"
#endif

static const char *	envoypath = NULL;
static char		ttypath[32];
static int		pt_pid = -1, pt_tty = -1, pt_pty = -1, pt_epid;

static int		ptrecv(int sock);
static void		ptterminak(int flags, int width, int height);
static void		ptslave(char **argv, int flags);
static int		ptenvoy(int release);
static void		ptreaper(int sock);
static RETSIGTYPE	catch_sig(int);
static void		install_handler(int, RETSIGTYPE (*handler)(int));

/*
 * External error handler
 */
extern void		error(const char *fmt, ...);

/*
 * ptsession
 *	start a session on a pty.
 *	argv
 *		argument vector for the command to be run
 *	flags
 *		various flags that modify the behavior
 *		PTSESS_CONSOLE
 *			use console mode
 *		PTSESS_8BIT
 *			8bit clean tty
 *		PTSESS_LOGINSHELL
 *			mark the shell as login shell
 *	width, height
 *		initial screen size for the slave tty
 *	returns
 *		the pid of the child process
 */
int
ptsession(char **argv, int flags, int width, int height)
{
	int	sock, i;

	if ((sock = ptenvoy(0)) < 0)
		return -1;

	/* Receive pty names and file descriptors from envoy */
	i = ptrecv(sock);

	/* Reap the envoy process */
	ptreaper(sock);

	if (i < 0) {
		close(sock);
		return -1;
	}

	fcntl(pt_pty, F_SETFL, O_NDELAY);

	/* Block all signals */
	for (i = 0; i < NSIG; i++)
		install_handler(i, catch_sig);
	install_handler(SIGCHLD, SIG_IGN);

	pt_pid = fork();
	if (pt_pid < 0) {
		error("cannot fork");
		return -1;
	}

	if (pt_pid == 0) {
		ptterminak(flags, width, height);
		ptslave(argv, flags); /* doesn't return */
	}

	return 0;
}

/*
 * Get the pid of the slave process
 */
int
ptgetpid(void)
{
	return pt_pid;
}

int
ptgetpty(void)
{
	return pt_pty;
}

int
ptgettty(void)
{
	return pt_tty;
}

/*
 * Clean up after session finished
 */
void
ptcleanup(void)
{
	int	sock, status;

	if (pt_pid < 0)
		return;

	/* Make sure child has gone */
	(void) waitpid(pt_epid, &status, 0);

	if ((sock = ptenvoy(1)) < 0)
		return;

	close(pt_tty);
	close(pt_pty);

	/* Reap the envoy process */
	ptreaper(sock);
}

/*
 * Receive tty name and file descriptors from envoy
 */
static int
ptrecv(int sock)
{
	struct { struct cmsghdr cmsg; int fds[2]; } control;
	struct msghdr	msg;
	struct iovec	iov;
	int		*fdp, flags = 0, success, cmsglen, len;

	msg.msg_name	= NULL;
	msg.msg_namelen	= 0;
	msg.msg_iov	= &iov;
	msg.msg_iovlen	= 1;
	msg.msg_control	= &control;
	msg.msg_controllen = sizeof(control);

	iov.iov_base	= ttypath;
	iov.iov_len	= sizeof(ttypath);

	cmsglen = sizeof(struct cmsghdr) + 2 * sizeof(int);

	success = 0;

	len = recvmsg(sock, &msg, flags);
	if (len <= 0) {
		error("recvmsg failed");
	} else if (0 /* (msg.msg_flags & (MSG_TRUNC|MSG_CTRUNC)) */
		|| msg.msg_controllen != cmsglen) {
		error("recvmsg: truncated message");
	} else {
		fdp = (int *) control.cmsg.cmsg_data;
		pt_tty = fdp[0];
		pt_pty = fdp[1];
		success = 1;
	}

	return success? 0 : -1;
}

static int
ptenvoy(int release)
{
	char	*argv[4];
	char	*sp;
	int	sv[2], pid, k;

	if (!envoypath) {
#ifdef DEBUG
		if (!(envoypath = getenv("UTMP_ENVOY")))
#endif
			envoypath = UTMP_ENVOY;
	}

	/* Fix up argv0 */
	argv[0] = (char *) envoypath;
	if ((sp = strrchr(envoypath, '/')) != NULL && sp[1])
		argv[0] = sp + 1;
	k = 1;
	if (release) {
		argv[k++] = "-y";
		argv[k++] = ttypath;
	}
	argv[k++] = NULL;

	if (socketpair(AF_UNIX, SOCK_STREAM, 0, sv) < 0) {
		error("can't create socket for utmpenvoy");
		return -1;
	}

	/* Fork envoy */
	pid = fork();
	if (pid < 0) {
		error("can't fork for utmpenvoy");
		return -1;
	}

	/* Parent: return one socket */
	if (pid != 0) {
		pt_epid = pid;
		close(sv[1]);
		return sv[0];
	}

	/* Child: close peer socket, attach to fd 0 */
	if (sv[1] != 0) {
		dup2(sv[1], 0);
		close(sv[1]);
	}
	close(sv[0]);

	execvp(envoypath, argv);
	error("cannot execute %s", envoypath);
	exit(1);
}

/*
 * Reap the envoy process, and close socket.
 */
static void
ptreaper(int sock)
{
	int	status;

	/* Reap the envoy process */
	(void) waitpid(pt_epid, &status, 0);
	close(sock);
	pt_epid = -1;
}

/*
 * Install signal handler
 */
static void
install_handler(int sig, RETSIGTYPE (*handler)(int))
{
#ifdef HAVE_SIGACTION
	struct sigaction	act;

	memset(&act, 0, sizeof(act));
	sigemptyset(&act.sa_mask);
	act.sa_handler = handler;
	sigaction(sig, &act, NULL);
#else
	signal(sig, handler);
#endif
}

static RETSIGTYPE
catch_sig(int sig)
{
	ptcleanup();
	install_handler(sig, SIG_DFL);
	kill(getpid(), sig);
}

/*  Terminal mode structures. */
#ifdef _HPUX_SOURCE
# define _TERMIOS_INCLUDED
# include <bsdtty.h>
#endif

#ifdef HAVE_TERMIO_H
# include <termio.h>
#endif

#ifdef HAVE_TERMIOS_H
# include <termios.h>
#endif

#if defined(VDISCRD) && !defined(VDISCARD)
# define VDISCARD	VDISCRD
#endif

#if defined(VWERSE) && !defined(VWERASE)
# define VWERASE	VWERSE
#endif

#ifndef __FreeBSD__
#ifndef CINTR
#ifndef _POSIX_VDISABLE
#define _POSIX_VDISABLE 0
#endif

#undef CTRL
#define CTRL(c) ((c) - '@')

#ifdef SVR4
#define CINTR 0177
#define CQUIT CTRL('U')
#define CERASE CTRL('H')
#else
#define CINTR CTRL('C')
#define CQUIT CTRL('\\')
#define CERASE 0177
#endif

#define CEOF CTRL('D')
#define CKILL CTRL('U')
#define CEOL _POSIX_VDISABLE
#define CEOL2 _POSIX_VDISABLE
#define CNSWTCH _POSIX_VDISABLE
#define CSTART CTRL('Q')
#define CSTOP CTRL('S')
#define CSUSP CTRL('Z');
#define CDSUSP _POSIX_VDISABLE
#define CRPRNT CTRL('R')
#define CFLUSH CTRL('O')
#define CWERASE CTRL('W')
#define CLNEXT CTRL('V')
#endif
#endif

#ifndef CEOL
#define CEOL _POSIX_VDISABLE
#endif

/* Terminal setup for slave tty */
static void
ptterminak(int flags, int width, int height)
{
	struct termios	ttmode;
	int		i;

#ifdef TIOCCONS
	if (flags & PTSESS_CONSOLE) {
		int             on = 1;

		if (ioctl(pt_tty, TIOCCONS, (unsigned char *) &on) == -1) {
			error("cannot open console");
			return;
		}
	}
#endif /* TIOCCONS */

#if defined(_HPUX_SOURCE) || defined (__Lynx__)
	for (i = 0; i < sysconf(_SC_OPEN_MAX); i++)
		if (i != pt_tty)
			close(i);
#else
	for (i = 0; i < getdtablesize(); i++)
		if (i != pt_tty)
			close(i);
#endif

	dup(pt_tty);
	dup(pt_tty);
	dup(pt_tty);

	if (pt_tty > 2)
		close(pt_tty);

	if (setsid() < 0)
		error("failed to set process group");
#if defined(TIOCSCTTY)
	ioctl(0, TIOCSCTTY, 0);
#endif
	{
		int pgrp = getpid();

		ioctl(0, TIOCSPGRP, (char *) &pgrp);
		setpgid(0, 0);
		close(open(ttypath, O_WRONLY, 0));
		setpgid(0, 0);
	}

	/* init of termios structure              */
#if defined (__FreeBSD__) || (__NetBSD__)
	ioctl(0, TIOCGETA, (char *) &ttmode);
#else
#if defined (_HPUX_SOURCE) || defined(__Lynx__)
	tcgetattr(0, &ttmode);
#else
	ioctl(0, TCGETS, (char *) &ttmode);
#endif
#endif

#if defined(_HPUX_SOURCE) || defined(__Lynx__)
	ttmode.c_iflag = BRKINT | IGNPAR | ICRNL | IXON;
	ttmode.c_lflag = ISIG | IEXTEN | ICANON | ECHO | ECHOE | ECHOK;
#else
	ttmode.c_iflag = BRKINT | IGNPAR | ICRNL | IXON | IMAXBEL;
	ttmode.c_lflag = ISIG | IEXTEN | ICANON | ECHO
				| ECHOE | ECHOK | ECHOCTL | ECHOKE;
#endif
	ttmode.c_oflag = OPOST | ONLCR;
	ttmode.c_cflag = B9600 | CS8 | CREAD;

	ttmode.c_cc[VEOF] = CEOF;
#ifdef ALPHA
	(unsigned) ttmode.c_cc[VEOL] = CEOL;
#else
	ttmode.c_cc[VEOL] = CEOL;
#endif
	ttmode.c_cc[VINTR] = CINTR;
	ttmode.c_cc[VQUIT] = CQUIT;
	ttmode.c_cc[VERASE] = CERASE;
	ttmode.c_cc[VKILL] = CKILL;
#if defined(_HPUX_SOURCE) || defined(__Lynx__)
	ttmode.c_cc[VSUSP] = CSWTCH;
#else
	ttmode.c_cc[VSUSP] = CSUSP;
#endif
#ifdef VDSUSP
	ttmode.c_cc[VDSUSP] = CDSUSP;
#endif
	ttmode.c_cc[VSTART] = CSTART;
	ttmode.c_cc[VSTOP] = CSTOP;
#ifdef VREPRINT
	ttmode.c_cc[VREPRINT] = CRPRNT;
#endif
#ifdef VDISCARD
	ttmode.c_cc[VDISCARD] = CFLUSH;
#endif
#ifdef VWERASE
	ttmode.c_cc[VWERASE] = CWERASE;
#endif
#ifdef VLNEXT
	ttmode.c_cc[VLNEXT] = CLNEXT;
#endif
#ifdef VSWTC
	ttmode.c_cc[VSWTC] = 0;
#endif

#ifdef VSWTCH
	ttmode.c_cc[VSWTCH] = 0;
#endif
#if VMIN != VEOF
	ttmode.c_cc[VMIN] = 1;
#endif
#if VTIME != VEOL
	ttmode.c_cc[VTIME] = 0;
#endif

	if (!(flags & PTSESS_8BIT))
		ttmode.c_cflag = B9600 | PARENB | CS7 | CREAD;
#if defined (__FreeBSD__) || (__NetBSD__)
	ioctl(0, TIOCSETA, (char *) &ttmode);
#else
#ifdef _HPUX_SOURCE
	tcsetattr(0, TCSANOW, &ttmode);
#else
	ioctl(0, TCSETS, (char *) &ttmode);
#endif
#endif
	/* Set winsize */
	{
		struct winsize	wsize;

		wsize.ws_row = (unsigned short) height;
		wsize.ws_col = (unsigned short) width;
		ioctl(pt_tty, TIOCSWINSZ, &wsize);
	}
}

/* Set up the pty slave, and run the command */
static void
ptslave(char **argv, int flags)
{
	/* the stuff inspired from rxvt-2.18: */
	/* command interpreter path */
	if (argv != NULL) {
		execvp(argv[0], (char **) argv);
		error("can't execute \"%s\"", argv[0]);
	} else {
		const char     *argv0, *shell;
		char           *base;

		if ((shell = getenv("SHELL")) == NULL || *shell == '\0')
			shell = "/bin/sh";

		base = strrchr(shell, '/');
		argv0 = (base ? base + 1 : shell);
		if (flags & PTSESS_LOGINSHELL) {
			char           *p;

			argv0 = p = (char *) malloc(strlen(argv0) + 2);
			*p = '-';
			strcpy(p + 1, argv0);
		}
		execlp(shell, argv0, NULL);
		error("can't execute \"%s\"", shell);
	}
	exit(1);
}
