/*
 * Compatibility routines for BSD systems
 *
 * Much of the code snarfed from KDE kvt, which in turn borrowed
 * it from various other apps.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef BSD

#include <stdio.h>
#include <paths.h>
#include <utmp.h>
#include <string.h>

#ifndef UT_LINESIZE
# define UT_LINESIZE		sizeof(((struct utmp *) 0)->ut_line)
#endif
#ifndef UT_IDSIZE
# define UT_IDSIZE		sizeof(((struct utmp *) 0)->ut_id)
#endif

#ifndef _PATH_UTMP
# define _PATH_UTMP		"/etc/utmp"
#endif
#ifndef _PATH_TTYTAB
# define _PATH_TTYTAB		"/etc/ttytab"
#endif

static FILE	*uf = NULL;
static int	uf_readwrite = 0;
static int	uf_reopen = 0;
static char	*utmp_path = NULL;

void
utmpname(const char *filename)
{
	if (utmp_path) {
		free(utmp_path);
		utmp_path = NULL;
	}
	if (filename) {
		utmp_path = (char *) malloc(strlen(filename) + 1);
		if (utmp_path)
			strcpy(utmp_path, filename);
	}
	uf_reopen = 1;
}

void
setutent(void)
{
	const char	*pathname;

	if (uf_reopen) {
		endutent();
		uf_reopen = 0;
	}

	if (uf != NULL) {
		rewind(uf);
		return;
	}

	uf_readwrite = 1;
	if ((uf = fopen(_PATH_UTMP, "r+")) == NULL) {
		uf = fopen(_PATH_UTMP, "r");
		uf_readwrite = 0;
	}
}

void
endutent(void)
{
	if (uf) {
		fclose(uf);
		uf = NULL;
	}
}

/*
 * Reconstruct the tty name from ut_id
 */
static const char *
id2ttyname(const char *id)
{
	static char	ttyname[8 + UT_IDSIZE + 1];

	/* Construct the /dev/ttyXX path name */
	strcpy(ttyname, "/dev/tty");
	strncpy(ttyname + 8, id, UT_IDSIZE);
	ttyname[8 + UT_IDSIZE] = '\0';

	return ttyname;
}

/*
 * Look up the tty name in /etc/ttytab and return a slot number that
 * can be used to access the utmp file. We cannot use ttyslot()
 * because the tty name is not that of fd 0.
 */
static int
id2offset(const char *id)
{
	static char	cache_id[UT_IDSIZE + 1];
	static int	cache_offset = -1;
	const char	*ttyname;
	char		buf[200], name[200];
	int		i, bol;
	FILE		*fs;

	if (cache_offset >= 0 && !strncmp(cache_id, id, UT_IDSIZE))
		return cache_offset;

	/* Construct the /dev/ttyXX path name */
	ttyname = id2ttyname(id);

	/* Try to locate the tty in TTYTAB */
	if ((fs = fopen(_PATH_TTYTAB,"r")) == NULL)
		return(-1);

	i   = 1;
	bol = 1; /* at beginning of line */
	while (fgets(buf, sizeof(buf) - 1, fs) != NULL) {
		buf[sizeof(buf) - 1] = '\0';
		if (bol && buf[0] != '#' && sscanf(buf, "%s", name) == 1) {
			if (strcmp(ttyname, name) == 0) {
				strncpy(cache_id, ttyname + 8, UT_IDSIZE);
				cache_offset = i * sizeof(struct utmp);
				fclose(fs);
				return (cache_offset);
			}
			i++;
		}
		bol = (strchr(buf, '\n') != NULL);
	}

	fclose(fs);
	return (-1);
}

/*
 * Get the next entry found in utmp file
 */
struct utmp *
getutent(void)
{
	static struct utmp	u;

	if (uf == NULL)
		setutent();
	if (uf && fread(&u, sizeof(u), 1, uf) == 1)
		return &u;
	return NULL;
}

/*
 * Search offset of the tty indicated by ut_id,
 * i.e. /dev/tty$id, and write out the (cleared) entry.
 *
 * Note: incomplete implementation; doesn't deal with RUN_LVL, BOOT_TIME
 * etc entries.
 */
void
pututline(struct utmp *u)
{
	int	offset, current, closeuf = 0;

	/* If ut_id not set, try to get it from the device name */
	if (u->ut_id[0] == '\0') {
		if (strncmp(u->ut_line, "tty", 3) || u->ut_line[3] == '\0')
			return;
		strncpy(u->ut_id, u->ut_line + 3, UT_IDSIZE);
	}

	/* Punt if not a valid tty name */
	if ((offset = id2offset(u->ut_id)) < 0)
		return;

	if (uf == NULL) {
		setutent();
		closeuf++;
	}

	if (uf != NULL && uf_readwrite) {
		current = ftell(uf);
		fseek(uf, offset, 0);
		fwrite((void *) u, sizeof(*u), 1, uf);
		fseek(uf, current, 0);
	}

	if (uf != NULL && closeuf)
		endutent();
}

#endif /* BSD */
