/*
 * Small setuid helper program for obtaining a pty pair
 *
 * When called in allocate mode, the program expects on fd 0 a UNIX socket pair
 *	for sending back the fd's of the open tty/pty to the caller.
 *	Along with these fds, the name of the tty and pty are sent.
 *
 * When called in release mode (-y option), the program expects the
 *	name of the tty to be released as an argument, and the descriptor
 *	of the pty on fd 1.
 *
 * Copyright (C) 1998, Olaf Kirch <okir@caldera.de>
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <grp.h>
#include <getopt.h>
#include <errno.h>
#include <syslog.h>

#include "utmp_funcs.h"

/* sgis have /dev/ptmx [bms] */
#ifdef __sgi__
# define SVR4
#endif

static int	getpty(char **ttypath, int *ttyfdp, int *ptyfdp);
static int	putpty(const char *);
static int	sendpty(const char *ttypath, int ttyfd, int ptyfd);
static void	usage(int exitcode);
static void	error(const char *fmt, ...);

int
main(int argc, char **argv)
{
	int	ttyfd, ptyfd;
	int	c, opt_yield = 0;
	char	*ttypath;

	if (geteuid() != 0) {
		error("%s: require effective root privileges", argv[0]);
		return 1;
	}

	/* Make sure we really have a socket */
	{
		struct sockaddr_un	un;
		int			ulen = sizeof(un);

		if (getpeername(0, (struct sockaddr *) &un, &ulen) < 0) {
			error("%s: can't get name of socket peer: %s",
				argv[0], strerror(errno));
			return 1;
		}
	}

	while ((c = getopt(argc, argv, "y:")) != -1) {
		switch (c) {
		case 'y':
			ttypath = optarg;
			opt_yield = 1;
			break;
		default:
			usage(1);
		}
	}

	if (optind < argc)
		usage(1);

	if (opt_yield) {
		if (putpty(ttypath) < 0)
			return 1;
		return 0;
	}

	if (getpty(&ttypath, &ttyfd, &ptyfd) < 0)
		return 1;

	if (sendpty(ttypath, ttyfd, ptyfd) < 0) {
		/* yieldpty(ptyfd); */
		return 1;
	}

	return 0;
}

static void
usage(int exitcode)
{
	fprintf(stderr, "usage: utmpenvoy [-y /dev/ptyxx]\n");
	exit(exitcode);
}

static int
getpty(char **ttypathp, int *ttyfdp, int *ptyfdp)
{
	int		ptyfd, ttyfd;
	unsigned char	*s3, *s4;
#ifndef SVR4
	static char	ptynam[32] = "/dev/ptyxx";
	static char	ttynam[32] = "/dev/ttyxx";
	static char     ptyc3[] = "pqrstuvwxyz";
	static char     ptyc4[] = "0123456789abcdef";
#else
	char		*ttynam, *ptsname();
#endif

	/*  First find a master pty that we can open.  
	 */
#ifdef SVR4
	ptyfd = open("/dev/ptmx", O_RDWR);
	if (ptyfd < 0) {
		error("Can't open a pseudo teletype");
		return (-1);
	}
	grantpt(ptyfd);
	unlockpt(ptyfd);
	fcntl(ptyfd, F_SETFL, O_NDELAY);
	ttynam = ptsname(ptyfd);
#else
	ptyfd = -1;
	for (s3 = ptyc3; ptyfd < 0 && *s3 != 0; s3++) {
		for (s4 = ptyc4; *s4 != 0; s4++) {
			ptynam[8] = ttynam[8] = *s3;
			ptynam[9] = ttynam[9] = *s4;
			if ((ptyfd = open(ptynam, O_RDWR)) >= 0) {
				if (geteuid() == 0
				 || access(ttynam, R_OK | W_OK) == 0) {
					break;
				} else {
					close(ptyfd);
					ptyfd = -1;
				}
			}
		}
	}
	if (ptyfd < 0) {
		error("Can't open a pseudo teletype");
		return (-1);
	}
	fcntl(ptyfd, F_SETFL, O_NDELAY);
#endif

	if ((ttyfd = open(ttynam, O_RDWR)) < 0) {
		error("could not open slave tty %s", ttynam);
		close(ptyfd);
		return (-1);
	}

#ifdef SVR4
	ioctl(ttyfd, I_PUSH, "ptem");
	ioctl(ttyfd, I_PUSH, "ldterm");
#endif

#ifndef SVR4
	{
		struct group	*gr;
		int		uid, gid;

		uid = getuid();
		if ((gr = getgrnam("tty")) != NULL)
			gid = gr->gr_gid;
		else
			gid = -1;
		fchown(ttyfd, uid, gid);
		fchmod(ttyfd, 0600);
	}
#endif

	/* stamp /etc/utmp */
	makeutent(&ttynam[5], getppid());

	*ttypathp = ttynam;
	*ptyfdp = ptyfd;
	*ttyfdp = ttyfd;
	return 0;
}

/*
 * Release the pty/tty pair
 * Note that we could also retrieve the tty name from the utmp
 * entry, but not all systems may actually _have_ utmp.
 */
static int
putpty(const char *ttypath)
{
	struct stat	fstb, pstb;

	if (strlen(ttypath) > sizeof(ttypath) - 1
	 || strncmp(ttypath, "/dev/tty", 8)) {
		error("PTY path %s: too long or bad location", ttypath);
		return -1;
	}
	if (stat(ttypath, &pstb) < 0) {
		error("cannot stat %s", ttypath);
		return -1;
	}
	if (fstat(1, &fstb) < 0) {
		error("cannot stat fd 1");
		return -1;
	}

	if (fstb.st_dev != pstb.st_dev || fstb.st_ino != pstb.st_ino) {
		error("Don't try cheating me: fd0 doesn't match %s", ttypath);
		return -1;
	}

	/* Clean up tty permissions. */
	chown(ttypath, 0, 0);

	/* Cleanup utmp */
	cleanutent(getppid());

	return 0;
}

/*
 * Send the tty pair over a UNIX socket pair (connected to fd 0)
 */
static int
sendpty(const char *ttypath, int ttyfd, int ptyfd)
{
	struct { struct cmsghdr	cmsg; int fds[2]; } control;
	struct msghdr	msg;
	struct iovec	iov[1];
	int		*fdp;

	msg.msg_name	= 0;
	msg.msg_namelen	= 0;
	msg.msg_iov	= iov;
	msg.msg_iovlen	= 1;
	msg.msg_control	= &control;
	msg.msg_controllen = sizeof(control);

	iov[0].iov_base	= (void *) ttypath;
	iov[0].iov_len	= strlen(ttypath) + 1;

	control.cmsg.cmsg_len = sizeof(struct cmsghdr) + 2 * sizeof(int);
	control.cmsg.cmsg_level = SOL_SOCKET;
	control.cmsg.cmsg_type = SCM_RIGHTS;
	fdp = (int *) &control.cmsg.cmsg_data;

	fdp[0] = ttyfd;
	fdp[1] = ptyfd;

	if (sendmsg(0, &msg, 0) < 0) {
		error("unable to send file descriptors");
		return -1;
	}

	return 0;
}

static void
error(const char *fmt, ...)
{
	va_list		ap;

	va_start(ap, fmt);
#ifdef HAVE_VSYSLOG
	vsyslog(LOG_WARNING, fmt, ap);
#else
	{
		char	overflowme[1024];

		setgid(getgid());
		setuid(getuid());
		vsprintf(overflowme, fmt, ap);
		syslog(LOG_WARNING, "%s", fmt);
	}
#endif
	va_end(ap);
}
